__version__ = "0.6.11"
__git_commiter_name__ = "Tal Yalon"
__git_commiter_email__ = "yalon@infinidat.com"
__git_branch__ = '(Not currently on any branch)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '2362168df63f6d6270eb53a778e034fa5efeec23'
__git_head_subject__ = "- added default field value by using default= - added kwargs to buffer's init so you can initialize fields in buffer from the ctor (MyBuffer(my_field=1))"
__git_head_message__ = ''
__git_dirty_diff__ = ''